@@ -31,7 +31,7 @@ import pycuda.driver as cuda
 from pytools import memoize
 from decorator import decorator
 import pycuda._driver as _drv
-import numpy
+import numpy as np
 
 
 
@@ -123,14 +123,14 @@ def _exact_div(dividend, divisor):
 
 def _int_ceiling(value, multiple_of=1):
     """Round C{value} up to be a C{multiple_of} something."""
-    # Mimicks the Excel "floor" function (for code stolen from occupany calculator)
+    # Mimicks the Excel "floor" function (for code stolen from occupancy calculator)
 
     from math import ceil
     return int(ceil(value/multiple_of))*multiple_of
 
 def _int_floor(value, multiple_of=1):
     """Round C{value} down to be a C{multiple_of} something."""
-    # Mimicks the Excel "floor" function (for code stolen from occupany calculator)
+    # Mimicks the Excel "floor" function (for code stolen from occupancy calculator)
 
     from math import floor
     return int(floor(value/multiple_of))*multiple_of
@@ -167,11 +167,10 @@ def get_default_device(default=0):
 
 
 
-def make_default_context():
-    ndevices = cuda.Device.count()
-    if ndevices == 0:
-        errmsg = "No CUDA enabled device found. Please check your installation."
-        raise RuntimeError, errmsg
+def make_default_context(ctx_maker=None):
+    if ctx_maker is None:
+        def ctx_maker(dev):
+            return dev.make_context()
 
     ndevices = cuda.Device.count()
     if ndevices == 0:
@@ -192,7 +191,7 @@ def make_default_context():
         except:
             pass
 
-    # If either CUDA_DEVICE or $HOME/.cuda_device is set, try to use it ;-)
+    # If either CUDA_DEVICE or $HOME/.cuda_device is set, try to use it
     if devn is not None:
         try:
             devn = int(devn)
@@ -201,7 +200,7 @@ def make_default_context():
                     " must be an integer")
 
         dev = cuda.Device(devn)
-        return dev.make_context()
+        return ctx_maker(dev)
 
     # Otherwise, try to use any available device
     else:
@@ -209,7 +208,7 @@ def make_default_context():
         for devn in xrange(ndevices):
             dev = cuda.Device(devn)
             try:
-                return dev.make_context()
+                return ctx_maker(dev)
             except cuda.Error:
                 pass
 
@@ -229,16 +228,26 @@ class DeviceData:
         self.max_threads = dev.get_attribute(drv.device_attribute.MAX_THREADS_PER_BLOCK)
         self.warp_size = dev.get_attribute(drv.device_attribute.WARP_SIZE)
 
-        if dev.compute_capability() < (1,2):
-            self.warps_per_mp = 24
-        else:
+        if dev.compute_capability() >= (2,0):
+            self.warps_per_mp = 48
+        elif dev.compute_capability() >= (1,2):
             self.warps_per_mp = 32
+        else:
+            self.warps_per_mp = 24
 
         self.thread_blocks_per_mp = 8
         self.registers = dev.get_attribute(drv.device_attribute.MAX_REGISTERS_PER_BLOCK)
         self.shared_memory = dev.get_attribute(drv.device_attribute.MAX_SHARED_MEMORY_PER_BLOCK)
 
-        self.smem_granularity = 16
+        if dev.compute_capability() >= (2,0):
+            self.smem_granularity = 128
+	else:
+            self.smem_granularity = 512
+
+        if dev.compute_capability() >= (2,0):
+            self.register_allocation_unit = "warp"
+	else:
+            self.register_allocation_unit = "block"
 
     def align(self, bytes, word_size=4):
         return _int_ceiling(bytes, self.align_bytes(word_size))
@@ -282,8 +291,13 @@ class OccupancyRecord:
 
         # copied literally from occupancy calculator
         alloc_warps = _int_ceiling(threads/devdata.warp_size)
-        alloc_regs = _int_ceiling(alloc_warps*2, 4)*16*registers
-        alloc_smem = _int_ceiling(shared_mem, 512)
+        alloc_smem = _int_ceiling(shared_mem, devdata.smem_granularity)
+	if devdata.register_allocation_unit == "warp":
+            alloc_regs = alloc_warps*32*registers
+	elif devdata.register_allocation_unit == "block":
+            alloc_regs = _int_ceiling(alloc_warps*2, 4)*16*registers
+	else:
+            raise ValueError("Improper register allocation unit:"+devdata.register_allocation_unit)
 
         if alloc_regs > devdata.registers:
             raise ValueError("too many registers")
@@ -331,53 +345,58 @@ def allow_user_edit(s, filename, descr="the file"):
 
 
 # C code generation helpers ---------------------------------------------------
-@memoize
-def platform_bits():
-    return tuple.__itemsize__ * 8
-
-
-
-
 def dtype_to_ctype(dtype, with_fp_tex_hack=False):
+    from pycuda.characterize import platform_bits
+    from sys import platform
+
     if dtype is None:
         raise ValueError("dtype may not be None")
 
-    import numpy
-    dtype = numpy.dtype(dtype)
-    if dtype == numpy.int64 and platform_bits() == 64:
-        return "long"
-    elif dtype == numpy.uint64 and platform_bits() == 64:
-        return "unsigned long"
-    elif dtype == numpy.int32:
+    dtype = np.dtype(dtype)
+    if dtype == np.int64 and platform_bits() == 64:
+        if 'win32' in platform:
+            return "long long"
+        else:
+            return "long"
+    elif dtype == np.uint64 and platform_bits() == 64:
+        if 'win32' in platform:
+            return "unsigned long long"
+        else:
+            return "unsigned long"
+    elif dtype == np.int32:
         return "int"
-    elif dtype == numpy.uint32:
+    elif dtype == np.uint32:
         return "unsigned int"
-    elif dtype == numpy.int16:
+    elif dtype == np.int16:
         return "short int"
-    elif dtype == numpy.uint16:
+    elif dtype == np.uint16:
         return "short unsigned int"
-    elif dtype == numpy.int8:
+    elif dtype == np.int8:
         return "signed char"
-    elif dtype == numpy.uint8:
+    elif dtype == np.uint8:
         return "unsigned char"
-    elif dtype == numpy.bool:
+    elif dtype == np.bool:
         return "bool"
-    elif dtype == numpy.float32:
+    elif dtype == np.float32:
         if with_fp_tex_hack:
             return "fp_tex_float"
         else:
             return "float"
-    elif dtype == numpy.float64:
+    elif dtype == np.float64:
         if with_fp_tex_hack:
             return "fp_tex_double"
         else:
             return "double"
-    elif dtype == numpy.complex64:
+    elif dtype == np.complex64:
         return "pycuda::complex<float>"
-    elif dtype == numpy.complex128:
+    elif dtype == np.complex128:
         return "pycuda::complex<double>"
     else:
-        raise ValueError, "unable to map dtype '%s'" % dtype
+        import pycuda.gpuarray as gpuarray
+        try:
+            return gpuarray.vec._dtype_to_c_name[dtype]
+        except KeyError:
+            raise ValueError, "unable to map dtype '%s'" % dtype
 
 
 
@@ -385,7 +404,7 @@ def dtype_to_ctype(dtype, with_fp_tex_hack=False):
 # C argument lists ------------------------------------------------------------
 class Argument:
     def __init__(self, dtype, name):
-        self.dtype = numpy.dtype(dtype)
+        self.dtype = np.dtype(dtype)
         self.name = name
 
     def __repr__(self):
@@ -406,7 +425,11 @@ class ScalarArg(Argument):
 
     @property
     def struct_char(self):
-        return self.dtype.char
+        result = self.dtype.char
+        if result == "V":
+            result = "%ds" % self.dtype.itemsize
+
+        return result
 
 
 
@@ -433,28 +456,37 @@ def parse_c_arg(c_arg):
     tp = c_arg[:decl_match.start()]
     tp = " ".join(tp.split())
 
-    if tp == "float": dtype = numpy.float32
-    elif tp == "double": dtype = numpy.float64
-    elif tp == "pycuda::complex<float>": dtype = numpy.complex64
-    elif tp == "pycuda::complex<double>": dtype = numpy.complex128
-    elif tp in ["int", "signed int"]: dtype = numpy.int32
-    elif tp in ["unsigned", "unsigned int"]: dtype = numpy.uint32
+    from pycuda.characterize import platform_bits
+    from sys import platform
+
+    if tp == "float": dtype = np.float32
+    elif tp == "double": dtype = np.float64
+    elif tp == "pycuda::complex<float>": dtype = np.complex64
+    elif tp == "pycuda::complex<double>": dtype = np.complex128
+    elif tp in ["int", "signed int"]: dtype = np.int32
+    elif tp in ["unsigned", "unsigned int"]: dtype = np.uint32
     elif tp in ["long", "long int"]:
-        if platform_bits() == 64:
-            dtype = numpy.int64
+        if platform_bits() == 64 and 'win32' not in platform:
+            dtype = np.int64
         else:
-            dtype = numpy.int32
-    elif tp in ["unsigned long", "unsigned long int"]:
-        if platform_bits() == 64:
-            dtype = numpy.uint64
+            dtype = np.int32
+    elif tp in ["unsigned long", "unsigned long int", "long unsigned int"]:
+        if platform_bits() == 64 and 'win32' not in platform:
+            dtype = np.uint64
         else:
-            dtype = numpy.uint32
-    elif tp in ["short", "short int"]: dtype = numpy.int16
-    elif tp in ["unsigned short", "unsigned short int"]: dtype = numpy.uint16
-    elif tp in ["char"]: dtype = numpy.int8
-    elif tp in ["unsigned char"]: dtype = numpy.uint8
-    elif tp in ["bool"]: dtype = numpy.bool
-    else: raise ValueError, "unknown type '%s'" % tp
+            dtype = np.uint32
+    elif tp in ["short", "short int"]: dtype = np.int16
+    elif tp in ["unsigned short", "unsigned short int", "short unsigned int"]: 
+        dtype = np.uint16
+    elif tp in ["char", "signed char"]: dtype = np.int8
+    elif tp in ["unsigned char"]: dtype = np.uint8
+    elif tp in ["bool"]: dtype = np.bool
+    else:
+        import pycuda.gpuarray as gpuarray
+        try:
+            dtype = gpuarray.vec._c_name_to_dtype[tp]
+        except KeyError:
+            raise ValueError("unknown type '%s'" % tp)
 
     return arg_class(dtype, name)
 
@@ -467,6 +499,11 @@ def get_arg_type(c_arg):
 
 
 
+context_dependent_memoized_functions = []
+
+
+
+
 @decorator
 def context_dependent_memoize(func, *args):
     try:
@@ -481,6 +518,7 @@ def context_dependent_memoize(func, *args):
     try:
         return ctx_dict[cur_ctx][args]
     except KeyError:
+        context_dependent_memoized_functions.append(func)
         arg_dict = ctx_dict.setdefault(cur_ctx, {})
         result = func(*args)
         arg_dict[args] = result
@@ -488,6 +526,18 @@ def context_dependent_memoize(func, *args):
 
 
 
+def clear_context_caches():
+    for func in context_dependent_memoized_functions:
+        try:
+            ctx_dict = func._pycuda_ctx_dep_memoize_dic
+        except AttributeError:
+            pass
+        else:
+            ctx_dict.clear()
+
+
+
+
 
 def mark_cuda_test(inner_f):
     def f(*args, **kwargs):
@@ -503,7 +553,12 @@ def mark_cuda_test(inner_f):
             inner_f(*args, **kwargs)
         finally:
             ctx.pop()
-            ctx.detach()
+
+            from pycuda.tools import clear_context_caches
+            clear_context_caches()
+
+            from gc import collect
+            collect()
 
     try:
         from py.test import mark as mark_test
@@ -511,3 +566,8 @@ def mark_cuda_test(inner_f):
         return f
 
     return mark_test.cuda(f)
+
+
+
+
+# vim: foldmethod=marker
